//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file ida.cpp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "ida.h"
#include "mappedcube.h"

extern unsigned short (*sSliceMove)[N_MOVE];
extern unsigned short (*cPosMove)[N_MOVE];
extern unsigned short(*twistMove)[N_MOVE];
extern unsigned short (*twistConj)[N_SYM];
extern SymRep (*flipSliceMove)[N_MOVE];
extern char moveConj[N_SYM][N_MOVE];
extern char symMult[N_SYM][N_SYM];

extern char* phase1Prun;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
IDA::IDA(MappedCube mc)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	n[0].axis=255;//invalid

	n[0].UDIdx=mc.flipSliceUD.classIdx;
	n[0].UDSym=mc.flipSliceUD.sym;
	n[0].UDTwist=mc.twistUD;

	n[0].RLIdx=mc.flipSliceRL.classIdx;
	n[0].RLSym=mc.flipSliceRL.sym;
	n[0].RLTwist=mc.twistRL;

	n[0].FBIdx=mc.flipSliceFB.classIdx;
	n[0].FBSym=mc.flipSliceFB.sym;
	n[0].FBTwist=mc.twistFB;

	n[1].axis=0;
	n[1].power=0;

	depth=1;
	nodeCount=0;
	idCount=0;
	cPos=mc.cPos;
	sSliceUD=mc.sSliceUD;
	sSliceRL=mc.sSliceRL;
	sSliceFB=mc.sSliceFB;

}

#if INLINE == 1
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int IDA::NextSolution(int maxDepth)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	unsigned short p;
	int r_depth=0;
	register Node* np=&n[depth];
	SymRep (*LflipSliceMove)[N_MOVE]=flipSliceMove;
	char (*LmoveConj)[N_MOVE]=moveConj;
	char (*LsymMult)[N_SYM]=symMult;
	unsigned short(*LtwistMove)[N_MOVE]=twistMove;
	unsigned short (*LtwistConj)[N_SYM]=twistConj;
	char* Lphase1Prun=phase1Prun;
	bool (IDA::*LIsIdCube)()=IsIdCube;
	unsigned int LNodeCount=nodeCount;

incPower:
	__asm
	{	
//		p=++np->power;
		mov ecx,dword ptr [np]
		mov dl,byte ptr [ecx.power]
		add	dl,1
		mov byte ptr [ecx.power],dl
		movzx dx,dl
		mov word ptr [p],dx
//if (p>MAXPOWER) goto incAxis;
		cmp dx,MAXPOWER
		jg incAxis
	}
	
turn:
	__asm
	{
//		LNodeCount++;
		mov eax,dword ptr [LNodeCount]
		add eax,1
		mov dword ptr [LNodeCount],eax



		push esi;we use esi for T and r_depth
		push edi;we use edi for nT
		
		mov ebx,dword ptr[np]

//		T= (p-1)*(MAXAXIS+1)+np->axis;
		mov         ecx,dword ptr [p]
		and         ecx,0FFFFh
		sub         ecx,1
		imul        ecx,ecx,MAXAXIS+1
		xor         eax,eax
		mov         al,byte ptr [ebx.axis]
		add         ecx,eax
		mov         esi,ecx; esi=T

//		sr=LflipSliceMove[(np-1)->UDIdx][LmoveConj[(np-1)->UDSym][T]];
		xor         edx,edx
		mov         dx,word ptr [ebx.UDIdx - TYPE node]; (np-1)->UDIdx
		imul        edx,edx,N_MOVE*(TYPE SymRep) 
		mov         eax,dword ptr [LflipSliceMove]
		add         eax,edx
		xor         edx,edx
		mov         dx,word ptr [ebx.UDSym - TYPE node]
		push edx
		
		imul        edx,edx,N_MOVE
		mov         ecx,dword ptr [LmoveConj]
		add         ecx,edx
		movsx       ecx,byte ptr [ecx+esi]
		mov         edx,dword ptr [eax+ecx*4]; edx=sr

//		np->UDIdx=sr.classIdx;
		mov word ptr [ebx.UDIdx],dx;

//		np->UDSym=LsymMult[sr.sym][(np-1)->UDSym];
		shr edx,0x10; sr.sym
		imul        edx,edx,N_SYM
		mov         eax,dword ptr [LsymMult]
		add         eax,edx
//		xor         edx,edx
//		mov         dx,word ptr [ebx.UDSym-TYPE node]
		pop edx
		movsx       ax,byte ptr [eax+edx]
		mov         word ptr [ebx.UDSym],ax

//		np->UDTwist=LtwistMove[(np-1)->UDTwist][T];
		xor         eax,eax
		mov         ax,word ptr [ebx.UDTwist-TYPE node]
		imul        eax,eax,N_MOVE*(TYPE short)
		mov         ecx,dword ptr [LtwistMove]
		add         ecx,eax
		mov         cx,word ptr [ecx+esi*2]
		mov         word ptr [ebx.UDTwist],cx

//		nT=LmoveConj[1][T];//nT=M_URF3*T*M_URF3^-1
		mov edx,dword ptr [LmoveConj]
		movsx edi,byte ptr [edx+esi+N_MOVE];nT

//		sr=LflipSliceMove[(np-1)->RLIdx][LmoveConj[(np-1)->RLSym][nT]];
		xor         edx,edx
		mov         dx,word ptr [ebx.RLIdx - TYPE node];
		imul        edx,edx,N_MOVE*(TYPE SymRep) 
		mov         eax,dword ptr [LflipSliceMove]
		add         eax,edx
		xor         edx,edx
		mov         dx,word ptr [ebx.RLSym - TYPE node]
		push edx

		imul        edx,edx,N_MOVE
		mov         ecx,dword ptr [LmoveConj]
		add         ecx,edx
		movsx       ecx,byte ptr [ecx+edi]
		mov         edx,dword ptr [eax+ecx*4]; edx=sr

//		np->RLIdx=sr.classIdx;
		mov word ptr [ebx.RLIdx],dx;

//		np->RLSym=LsymMult[sr.sym][(np-1)->RLSym];
		shr edx,0x10; sr.sym
		imul        edx,edx,N_SYM
		mov         eax,dword ptr [LsymMult]
		add         eax,edx
//		xor         edx,edx
//		mov         dx,word ptr [ebx.RLSym-TYPE node]
		pop edx
		movsx       ax,byte ptr [eax+edx]
		mov         word ptr [ebx.RLSym],ax

//		np->RLTwist=LtwistMove[(np-1)->RLTwist][nT];
		xor         eax,eax
		mov         ax,word ptr [ebx.RLTwist-TYPE node]
		imul        eax,eax,N_MOVE*(TYPE short)
		mov         ecx,dword ptr [LtwistMove]
		add         ecx,eax
		mov         cx,word ptr [ecx+edi*2]
		mov         word ptr [ebx.RLTwist],cx

//		nT=LmoveConj[2][T];//nT=M_URF3^-1*T*M_URF3
		mov edx,dword ptr [LmoveConj]
		movsx edi,byte ptr [edx+esi+N_MOVE*2];nT

//		sr=LflipSliceMove[(np-1)->FBIdx][LmoveConj[(np-1)->FBSym][nT]];
		xor         edx,edx
		mov         dx,word ptr [ebx.FBIdx - TYPE node];
		imul        edx,edx,N_MOVE*(TYPE SymRep) 
		mov         eax,dword ptr [LflipSliceMove]
		add         eax,edx
		xor         edx,edx
		mov         dx,word ptr [ebx.FBSym - TYPE node]
		push edx
		imul        edx,edx,N_MOVE
		mov         ecx,dword ptr [LmoveConj]
		add         ecx,edx
		movsx       ecx,byte ptr [ecx+edi]
		mov         edx,dword ptr [eax+ecx*4]; edx=sr

//		np->FBIdx=sr.classIdx;
		mov word ptr [ebx.FBIdx],dx;

//		np->FBSym=LsymMult[sr.sym][(np-1)->FBSym];
		shr edx,0x10; sr.sym
		imul        edx,edx,N_SYM
		mov         eax,dword ptr [LsymMult]
		add         eax,edx
//		xor         edx,edx
//		mov         dx,word ptr [ebx.FBSym-TYPE node]
		pop edx
		movsx       ax,byte ptr [eax+edx]
		mov         word ptr [ebx.FBSym],ax

//		np->FBTwist=LtwistMove[(np-1)->FBTwist][nT];
		xor         eax,eax
		mov         ax,word ptr [ebx.FBTwist-TYPE node]
		imul        eax,eax,N_MOVE*(TYPE short)
		mov         ecx,dword ptr [LtwistMove]
		add         ecx,eax
		mov         cx,word ptr [ecx+edi*2]
		mov         word ptr [ebx.FBTwist],cx

		pop edi
		pop esi

//		if (r_depth>MAXPRUN) goto right;
		mov esi,dword ptr [r_depth]
		cmp esi,MAXPRUN
		jg	right

//		idx=LtwistConj[np->UDTwist][np->UDSym]+N_CTWIST*np->UDIdx;
//		mov ebx,dword ptr [np]
		xor ecx,ecx
		mov cx,word ptr [ebx.UDTwist];np->UDTwist
		imul ecx,ecx,N_SYM*(TYPE short); offset
		mov edx,dword ptr [LtwistConj]
		add ecx,edx; &LtwistConj[np->UDTwist]
		xor eax,eax;
		mov ax,word ptr [ebx.UDSym]
		xor edx,edx
		mov dx,word ptr [ecx+eax*2]
		xor ecx,ecx
		mov cx,word ptr [ebx.UDIdx]
		imul ecx,ecx,N_CTWIST
		add edx,ecx; edx=idx

//		tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
//			(char)((Lphase1Prun[idx>>1]&0xf0)>>4));

		sar edx,1; low bit to carry for jc
		mov ecx,dword ptr [Lphase1Prun]
		mov al,byte ptr [ecx+edx]
		jc got_nibble1  ; low bit clear? skip
		sar eax,4       ; get hi nibble
got_nibble1:
		and eax,0Fh

//		if (tag>r_depth+1) goto incAxis;
		mov         ecx,esi;esi=r_depth
		add         ecx,1
		cmp         eax,ecx
		jg			incAxis
//		if (tag>r_depth) goto incPower;
		cmp eax,esi
		jg	incPower

//		idx=LtwistConj[np->RLTwist][np->RLSym]+N_CTWIST*np->RLIdx;
		xor ecx,ecx
		mov cx,word ptr [ebx.RLTwist]
		imul ecx,ecx,N_SYM*(TYPE short); offset
		mov edx,dword ptr [LtwistConj]
		add ecx,edx; &LtwistConj[np->RLTwist]
		xor eax,eax;
		mov ax,word ptr [ebx.RLSym]
		xor edx,edx
		mov dx,word ptr [ecx+eax*2]
		xor ecx,ecx
		mov cx,word ptr [ebx.RLIdx]
		imul ecx,ecx,N_CTWIST
		add edx,ecx; edx=idx

//		tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
//			(char)((Lphase1Prun[idx>>1]&0xf0)>>4));

		sar edx,1; low bit to carry for jc
		mov ecx,dword ptr [Lphase1Prun]
		mov al,byte ptr [ecx+edx]
		jc got_nibble2  ; low bit clear? skip
		sar eax,4       ; get hi nibble
got_nibble2:
		and eax,0Fh

//		if (tag>r_depth+1) goto incAxis;
		mov         ecx,esi;esi=r_depth
		add         ecx,1
		cmp         eax,ecx
		jg			incAxis
//		if (tag>r_depth) goto incPower;
		cmp eax,esi
		jg	incPower

//		idx=LtwistConj[np->FBTwist][np->FBSym]+N_CTWIST*np->FBIdx;
		xor ecx,ecx
		mov cx,word ptr [ebx.FBTwist]
		imul ecx,ecx,N_SYM*(TYPE short); offset
		mov edx,dword ptr [LtwistConj]
		add ecx,edx; &LtwistConj[np->FBTwist]
		xor eax,eax;
		mov ax,word ptr [ebx.FBSym]
		xor edx,edx
		mov dx,word ptr [ecx+eax*2]
		xor ecx,ecx
		mov cx,word ptr [ebx.FBIdx]
		imul ecx,ecx,N_CTWIST
		add edx,ecx; edx=idx

//		tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
//			(char)((Lphase1Prun[idx>>1]&0xf0)>>4));

		sar edx,1; low bit to carry for jnc
		mov ecx,dword ptr [Lphase1Prun]
		mov al,byte ptr [ecx+edx]
		jc got_nibble3  ; low bit clear? skip
		sar eax,4       ; get hi nibble
got_nibble3:
		and eax,0Fh

//		if (tag>r_depth+1) goto incAxis;
		mov         ecx,esi;esi=r_depth
		add         ecx,1
		cmp         eax,ecx
		jg			incAxis
//		if (tag>r_depth) goto incPower;
		cmp eax,esi
		jg	incPower

//		if (r_depth==0)
		cmp esi,0
		jne right

//      if (!IsIdCube()) goto incPower;
		mov ecx, dword ptr(this); may be compiler-specific!!!!
		call LIsIdCube	
		and eax,0FFh
		test eax,eax
		je incPower
	}
	nodeCount=LNodeCount;
	return depth;

right:
	__asm
	{
//		r_depth--;
		sub esi,1
		mov dword ptr [r_depth],esi

//		np++; np->axis=0;
		add ebx,TYPE node
		mov dword ptr [np],ebx
		mov byte ptr [ebx.axis],0

//		goto checkNeighbourAxis;
		jmp checkNeighbourAxis;
	}

incAxis:
	__asm
	{
//		if (++np->axis>MAXAXIS) goto left;
		mov edx,dword ptr [np]
		mov al,byte ptr [edx.axis]
		add al,1
		mov byte ptr [edx.axis],al
		cmp al,MAXAXIS
		jg left
	}

checkNeighbourAxis:
	__asm
	{
//		if (np->axis==(np-1)->axis) goto incAxis;
		mov edx,dword ptr [np]
		mov al,byte ptr [edx.axis]
		mov cl,byte ptr [edx.axis-TYPE node]
		cmp cl,al
		je incAxis

//		if (np->axis+3==(np-1)->axis) goto incAxis;
		add al,3
		cmp cl,al
		je incAxis

//		np->power=1;
		mov byte ptr [edx.power],1

//		p=1;
		mov word ptr [p],1
		jmp turn
	}

left:
	r_depth++;
	if (depth-r_depth==0)
	{	cout<<"Depth "<<depth<<" completed. "<<LNodeCount<<" nodes, "<<idCount<< " idchecks."<<endl;
		LNodeCount=idCount=0;
		if (depth==maxDepth)
		{
			nodeCount=LNodeCount;
			return -1; //no more solution
		}
		depth++;
		np->axis=0; np->power=1; p=1;
		goto turn;
	}
	np--; goto incPower;
}
#else
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int IDA::NextSolution(int maxDepth)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	register Node* np=&n[depth];
	SymRep (*LflipSliceMove)[N_MOVE]=flipSliceMove;
	char (*LmoveConj)[N_MOVE]=moveConj;
	char (*LsymMult)[N_SYM]=symMult;
	unsigned short(*LtwistMove)[N_MOVE]=twistMove;
	unsigned short (*LtwistConj)[N_SYM]=twistConj;
	char* Lphase1Prun=phase1Prun;
	bool (IDA::*LIsIdCube)()=IsIdCube;
	int LNodeCount=nodeCount;

	int idx,r_depth=0,tag;
	unsigned short p,T,nT;
	SymRep sr;

incPower:
	p=++np->power;
	if (p>MAXPOWER) goto incAxis;
	
turn:
	LNodeCount++;
	T= (p-1)*(MAXAXIS+1)+np->axis;

	sr=LflipSliceMove[(np-1)->UDIdx][LmoveConj[(np-1)->UDSym][T]];//c*T=s^-1*c*s
	np->UDIdx=sr.classIdx;
	np->UDSym=LsymMult[sr.sym][(np-1)->UDSym];//s*s=s
	np->UDTwist=LtwistMove[(np-1)->UDTwist][T];


	nT=LmoveConj[1][T];//nT=M_URF3*T*M_URF3^-1
	sr=LflipSliceMove[(np-1)->RLIdx][LmoveConj[(np-1)->RLSym][nT]];
	np->RLIdx=sr.classIdx;
	np->RLSym=LsymMult[sr.sym][(np-1)->RLSym];
	np->RLTwist=LtwistMove[(np-1)->RLTwist][nT];

	nT=LmoveConj[2][T];//nT=M_URF3^-1*T*M_URF3
	sr=LflipSliceMove[(np-1)->FBIdx][LmoveConj[(np-1)->FBSym][nT]];
	np->FBIdx=sr.classIdx;
	np->FBSym=LsymMult[sr.sym][(np-1)->FBSym];
	np->FBTwist=LtwistMove[(np-1)->FBTwist][nT];

	if (r_depth>MAXPRUN) goto right;//pruning not possible

	idx=LtwistConj[np->UDTwist][np->UDSym]+N_CTWIST*np->UDIdx;
	tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
		(char)((Lphase1Prun[idx>>1]&0xf0)>>4));
	if (tag>r_depth+1) goto incAxis;
	if (tag>r_depth) goto incPower;

	idx=LtwistConj[np->RLTwist][np->RLSym]+N_CTWIST*np->RLIdx;
	tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
		(char)((Lphase1Prun[idx>>1]&0xf0)>>4));
	if (tag>r_depth+1) goto incAxis;
	if (tag>r_depth) goto incPower;

	idx=LtwistConj[np->FBTwist][np->FBSym]+N_CTWIST*np->FBIdx;
	tag=(idx%2==1?Lphase1Prun[idx>>1]&(char)0x0f:
		(char)((Lphase1Prun[idx>>1]&0xf0)>>4));
	if (tag>r_depth+1) goto incAxis;
	if (tag>r_depth) goto incPower;

	if (r_depth==0)
	{
		if (!IsIdCube()) goto incPower;
		nodeCount=LNodeCount;
		return depth;//solution found
	}

right:
	r_depth--;
	np++; np->axis=0;
	goto checkNeighbourAxis;

incAxis:
	if (++np->axis>MAXAXIS) goto left;

checkNeighbourAxis:
	if (np->axis==(np-1)->axis) goto incAxis;//UU etc. not allowed
	if (np->axis+3==(np-1)->axis) goto incAxis;//UD etc. not allowed
	np->power=1;
	p=1;
	goto turn;

left:
	r_depth++;
	if (depth-r_depth==0)
	{	cout<<"Depth "<<depth<<" completed. "<<LNodeCount<<" nodes, "<<idCount<< " idchecks."<<endl;
		LNodeCount=idCount=0;
		if (depth==maxDepth) return -1; //no more solution
		depth++;
		np->axis=0; np->power=1; p=1;
		goto turn;
	}
	np--; goto incPower;
}
#endif

bool IDA::IsIdCube()
{
	idCount++;
	int power;
	unsigned char T;
	Node* np=&n[1];

	unsigned short test=cPos;
	for (int i=1;i<=depth;i++)
	{	
		power=np->power;
		T=np++->axis;
		for (int j=0;j<power;j++)
		{
			test=cPosMove[test][T];
		}
	}
	if (test!=0) return false;

	np=&n[1];
	test=sSliceUD;
	{for (int i=1;i<=depth;i++)
	{	
		power=np->power;
		T=np++->axis;
		for (int j=0;j<power;j++)
		{
			test=sSliceMove[test][T];
		}
	}}
	if (test!=0) return false;

	np=&n[1];
	test=sSliceRL;
	{for (int i=1;i<=depth;i++)
	{	
		power=np->power;
		T=moveConj[1][np++->axis];
		for (int j=0;j<power;j++)
		{
			test=sSliceMove[test][T];
		}
	}}
	if (test!=0) return false;

	np=&n[1];
	test=sSliceFB;
	{for (int i=1;i<=depth;i++)
	{	
		power=np->power;
		T=moveConj[2][np++->axis];
		for (int j=0;j<power;j++)
		{
			test=sSliceMove[test][T];
		}
	}}
	if (test!=0) return false;

	return true;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void IDA::PrintManeuver(ostream& ostr)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	for (int i=1;i<=depth;i++)
	{	switch (n[i].axis)
		{	case U1: ostr<<"U";break;
			case R1: ostr<<"R";break;
			case F1: ostr<<"F";break;
			case D1: ostr<<"D";break;
			case L1: ostr<<"L";break;
			case B1: ostr<<"B";break;
		}
		switch (n[i].power)
		{	case 1: ostr<<" ";break;
			case 2: ostr<<"2 ";break;
			case 3: ostr<<"' ";break;
		}
	}
	ostr<<" ("<<depth<<"f*)"<<endl;
	ostr.flush();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void IDA::PrintInvManeuver(ostream& ostr)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	for (int i=depth;i>0;i--)
	{	switch (n[i].axis)
		{	case U1: ostr<<"U";break;
			case R1: ostr<<"R";break;
			case F1: ostr<<"F";break;
			case D1: ostr<<"D";break;
			case L1: ostr<<"L";break;
			case B1: ostr<<"B";break;
		}
		switch (n[i].power)
		{	case 1: ostr<<"' ";break;
			case 2: ostr<<"2 ";break;
			case 3: ostr<<" ";break;
		}
	}
	ostr<<" ("<<depth<<"f*)"<<endl;
	ostr.flush();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void IDA::PrintSeparation(ostream& ostr)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	ostr<<"---------------------------------------------------------------"<<endl;
}


