//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file ida.h

//The class IDA is to perform the iterative deepening A* with a lowerbound
//heuristic function (IDA*) to find the shortest maneuver to solve a cube.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#ifndef IDA_H
#define IDA_H

#include "constants.h"
#include "mappedcube.h"
#include <iostream.h>

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
typedef struct node
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	unsigned char axis;
	unsigned char power;

	unsigned short UDIdx;
	unsigned short UDSym;
	unsigned short UDTwist;

	unsigned short RLIdx;
	unsigned short RLSym;
	unsigned short RLTwist;
	
	unsigned short FBIdx;
	unsigned short FBSym;
	unsigned short FBTwist;
} Node;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class IDA
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	public:
	Node n[MAXSOLVERLENGTH+1];
	int depth;
	unsigned long nodeCount; 
	unsigned long idCount;

	unsigned short cPos;
	unsigned short sSliceUD;
	unsigned short sSliceRL;
	unsigned short sSliceFB;
	
	IDA(MappedCube);
	int NextSolution(int maxLength);
	bool IsIdCube();

	void PrintManeuver(ostream&);
	void PrintInvManeuver(ostream&);
	void PrintSeparation(ostream&);
};

#endif
