//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file mappedcube.h
//
//The an object of class MappedCube represents a cube in an abstract form, where
//the state of the cube is defined by a set of integer or SymRep coordinates.
//
//Class SymRep is needed when the conjugation by cube symmetries is used to define
//equivalence classes of coordinates. classIndex identifies the equivalence class
//and sym the symmetry.
//
//void MappedCube::InitXXXXXMoveTable() creates the tables which map the
//coordinates to the new coordinates when a face-turn is performed.
//void MappedCube::InitTwistConjugateTable() is needed for transforming the
//corner twist coordinate by a conjugation with a cube symmetry.
//void MappedCube::InitFlipSliceTwistPrunTable() creates the pruning table
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#include <iostream.h>


#ifndef MAPPEDCUBE_H
#define MAPPEDCUBE_H

#include "oriperm.h"

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SymRep
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	public:
	unsigned short classIdx;
	unsigned short sym;
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MappedCube
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  
	public:
	MappedCube(COPerm,EOPerm);
	MappedCube(int flipSliceTwist);
	MappedCube(unsigned char* generatorAxis, unsigned char* generatorPower);
	MappedCube& operator=(const MappedCube& mc);
	void Turn(Move T);
	void Print();
	int FlipSliceTwist();
	static SymRep FlipSliceSymRep(EOPerm);


//coodinates used for pruning
	SymRep flipSliceUD;
	SymRep flipSliceRL;
	SymRep flipSliceFB;
	unsigned short twistUD;
	unsigned short twistRL;
	unsigned short twistFB;

//other coordinates
	unsigned short cPos;
	unsigned short sSliceUD;
	unsigned short sSliceRL;
	unsigned short sSliceFB;

//initialize the tables
	static void InitFlipSliceMoveTable();
	static void InitTwistMoveTable();
	static void InitSortedSliceMoveTable();
	static void InitCPosMoveTable();

	static void InitTwistConjugateTable();
	static void InitFlipSliceTwistPrunTable();
	
};

#endif