//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file permutation.cpp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "permutation.h"
#include <iostream.h>
#include <iomanip.h>

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm::Perm(char sz):size(sz)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	data = new char[sz];
	for (char i=0;i<sz;i++) data[i]=i;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm::Perm(char sz, int index):size(sz)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	data = new char[sz];
	char* order = new char[sz-1];
	for (char m=0;m<sz;m++) data[m]=m;
	char buf;
	for (char i=2;i<=sz;i++)
	{	order[sz-i]=(char)(index%i);
		index /=i;
	}
	for (char k=0;k<sz-1;k++)
	{	for (char j=order[k];j>0;j--)
		{	buf =data[k+j];
			data[k+j]=data[k+j-1];
			data[k+j-1]=buf;
		}
	}
	delete [] order;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm::Perm(char sz,char* dat):size(sz)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  data = new char[sz];
	for (char j=0;j<sz;j++)
		data[j]=dat[j];
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm::Perm(const Perm& p):size(p.size)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	data = new char[size];
	for (char i=0;i<size;i++) data[i]=p.data[i];
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm& Perm::operator=(const Perm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  if (size==p.size)			//we want to define = also in the case
	for (char i=0;i<size;i++)	//we have different sizes for the arguments
		data[i]=p.data[i];
	else if (size>p.size)
	{	for (char i=0;i<p.size;i++)
			data[i]=p.data[i];
		for (char j=p.size;j<size;j++)
			data[j]=j;
	}
	else
	{	for (char k=0;k<size;k++)
		{  char t=k;
			while (p.data[t]>=size) t=p.data[t];
			data[k]=p.data[t];
		}
	}
	return *this;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm::~Perm()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	delete[] data;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int Perm::Index()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int s=0;
	for (char i=0;i<size-1;i++)
	{	int o=0;
		for (char j=char(i+1);j<size;j++)
			o += data[j]<data[i];
		s += o;
		s *= size-1-i;
	}
	return s;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm Perm::Inverse()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  Perm p(size);
	for (char i=0;i<size;i++) p.data[data[i]]=i;
	return p;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
bool Perm::Even()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int counter=0;
	for (char i=0;i<size-1;i++)
	{	int o=0;
		for (char j=char(i+1);j<size;j++)
			o += data[j]<data[i];
		counter += o;
	}
	if (counter%2==0) return 1;
	else return 0;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm Perm::operator*(const Perm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  char psize=p.size;
	if (size==psize)
	{	Perm result(size);
		for (char i=0;i<size;i++)
			result.data[i]=p.data[data[i]];
		return result;
	}
	else if (size<psize)
	{	Perm result(psize),q(psize);
		q=*this;
		for (char i=0;i<psize;i++)
			result.data[i]=p.data[q.data[i]];
		return result;
	}
	else if (size>psize)
	{	Perm result(size),q(size);
		q=p;
		for (char i=0;i<size;i++)
			result.data[i]=q.data[data[i]];
		return result;
	}
	return *this;//should never be called
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void Perm::Print()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	for (int i=0;i<size;i++) cout<<setw(3)<<(int)data[i]<<" ";
	cout<<endl;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
CPerm::CPerm():Perm(N_CORNER)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
CPerm::CPerm(int state):Perm(N_CORNER,state)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
CPerm::CPerm(char* dat):Perm(N_CORNER,dat)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm& CPerm::operator=(const Perm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	return Perm::operator=(p);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EPerm::EPerm():Perm(N_EDGE)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EPerm::EPerm(int state):Perm(N_EDGE,state)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EPerm::EPerm(char* dat):Perm(N_EDGE,dat)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Perm& EPerm::operator=(const Perm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	return Perm::operator=(p);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int Cnk(char n, char k)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int s;
 char i,j;
 if (n<k)   return 0;
 if (k>n/2)  k=(char)(n-k);
 for (s=1,i=n,j=1; i!=n-k; i--,j++)
 {  s *= i; s /= j;
 }
 return s;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int Fak(char n)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int s=1;
	for (char k=2;k<=n;k++)
		s *=k;
	return s;
}
